'''
These are the default hyperparameters for the algorithm. They're stored in a separate namespace for users who prefer to use default values
'''
## Population size
N_POP = 100  
## Bitstring length      
N_LEN = 64         
## Crossover rate     
R_CROSS = 0.5    
## Bit mutation rate       
R_MUT = 1.0 / N_LEN
## Number of generations      
N_GEN = 100             


# Getters (return float)

## Get population
def getPop():
    return N_POP
## Get bitsting length
def getLen():
    return N_LEN
## Get crossover rate
def getCross():
    return R_CROSS
## Get mutation rate
def getMut(): 
    return R_MUT
## Get number of generations
def getGen():
    return N_GEN



# Setters (return bool)

## Set population
def setPop(x):
    N_POP = x
    return True
## Set bitstring length
def setLen(x):
    N_LEN = x
    return True
## Set crossover rate
def setCross(x):
    R_CROSS = x
    return True
## Set mutation rate
def setMut(x):
    R_MUT = x
    return True
## Set number of generations
def setGen(x):
    N_GEN = x
    return True







'''
Functions used in the genetic search algorithm
'''

import numpy
import genSearch.interface as itf

## Generate Population of n_pop DNA-bitstings of n_len
def population(n_pop, n_len):
    return ["".join([str(s) for s in numpy.random.randint(0, 2, n_len)]) for _ in range(n_pop)]


## Index-match cost scores to population members
def fitness(population):
    return [gcost(c) for c in population]

## Return the fitness score of a DNA-bitstring 
def gcost(indv):
    z = itf.decoder(indv)
    return itf.cost(z)



## Run n_pop tournaments between k randomly selected individuals
def selection(pop, cost, k=3):
    return [tournament(pop, cost, k) for _ in range(len(pop))]

## Simulate a tournament between k individuals 
def tournament(pop, cost, k=3):
    # choose random winner
    v_ix = numpy.random.randint(len(pop))
    # competition
    for ix in numpy.random.randint(0, len(pop)-1, k-1):
        if cost[ix] < cost[v_ix]:
            v_ix = ix
    return pop[v_ix]
    

## Cross-over two DNA-bitstrings  
def crossover(p1, p2, r_cross):
    # define children
    c1, c2 = p1, p2
    # determine if a crossover happens
    if numpy.random.rand() < r_cross:
        # random crossover point
        pt = numpy.random.randint(0, len(c1))
        # crossover
        c1 = p1[:pt] + p2[pt:]
        c1 = p1[pt:] + p2[:pt]
    return c1, c2


## Mutate a DNA-bitstring
def mutate(indv, r_mut):
    f_indv = ""
    for i in range(len(indv)):
        if numpy.random.rand() < r_mut:
            f_indv += str((int(indv[i]) + 1) % 2)
        else:
            f_indv += indv[i]
    return f_indv 


## Return best solution and its cost
def best(pop, cost):
    lo_cost = min(cost)
    lo_ix = cost.index(lo_cost)
    return [itf.decoder(pop[lo_ix]), lo_cost]






'''
 The following functions are defined by the user and inputs into the genetic algorithm. These are default values which are later set to the user's respective functions.
'''



## Map objects to real numbers based on fitness criteria
def cost(object):
    return 1

## Establish a bijection between DNA-bitstrings and objects
def decoder(DNA):
    return 1



## @mainpage Primitive Genetic Algorithm
#
# @section intro_sec Overview
#
# This is a package definining a simple genetic algorithm. This algorithm finds optimum solutions to a problem by modeling evolutionary competition. Users establish a solution's 
# fitness via a `cost` function; the function maps inputs to real numbers based on user criteria. Users give a `decoder` function
# to establish an isomorphism between solutions and bit-strings. The algorithm generates a population of random solutions (as bit-strings), and 
# evaluates their individual fitnesses. Best solutions are selected via tournaments to generate a new population via processes analogous
# to sexual reproduction. This process is iterated a given number of times and the final best solution is returned. 
# \n
# \n
# This particular algorithm is primitive as it treats objects as  binary-strings; other genetic algorithms may abstract objects to more
# complex data structures. 
#

from genSearch.algorithms import *
from genSearch.interface import *
from genSearch.constants import *


from genSearch.helperFunctions import *
import genSearch.interface as itf



##
#
# @section intro_sec Overview
# This is the genetic algorithm. Users establish a bijection between bistrings and objects in `decoder`. Users establish a map from objects to real
# numbers in `cost`. The algorithm then runs and returns an approximation of the minimum of `cost`: the algorithm generates a pseudorandom population
# of bitstrings. It ranks them according to their corresponding object's cost. It selects a new population from these by running tournaments. These are 
# then crossed over. The offspring are then mutated. This process is iterated on the new population a given amount of times. When the iterations end, 
# the algorithm returns the fittest from the final generation
#
# @param Decoder: f: str -> <E>, internal binary strings to client-side solutions objects
# @param Cost: f: <E> -> Double,  client-side solution objects to real numbers
# @param n_pop Size of population
# @param n_len Length of bitstring
# @param r_cross Probability of crossover during reproduction stage
# @param r_mut Probability of a bit swapping during mutation stage
# @param n_gen Number of generations or iterations
# 
# @return solution: <E> type object
def genetic_simulation(decoder, cost):
    
    # configure interface
    itf.cost = cost
    itf.decoder = decoder
    
    # initialize population
    pop = population(cts.N_POP, cts.N_LEN)
    # iterate generations
    for gen in range(cts.N_GEN):
        # assign fitness scores
        fit = fitness(pop)
        sol = best(pop, fit)
        # select reproducing individuals
        vic = selection(pop, fit)
        # initialize next generation
        nxt_pop = []
        # reproduce
        for ps in range(0, cts.N_POP, 2):
            c1, c2 = crossover(vic[ps], vic[ps+1], cts.R_CROSS)
            c1, c2 = mutate(c1, cts.R_MUT), mutate(c2, cts.R_MUT)
            nxt_pop.append(c1)
            nxt_pop.append(c2)
        # update population
        pop = nxt_pop
    return sol